package app

import (
	"errors"
	"fmt"
	"os"
	"os/exec"
	"path"
	"strings"

	"coopcloud.tech/abra/cli/internal"
	"coopcloud.tech/abra/pkg/autocomplete"
	"coopcloud.tech/abra/pkg/client"
	"coopcloud.tech/abra/pkg/config"
	"coopcloud.tech/abra/pkg/runtime"
	"github.com/sirupsen/logrus"
	"github.com/urfave/cli"
)

var appCmdCommand = cli.Command{
	Name:    "command",
	Aliases: []string{"cmd"},
	Usage:   "Run app commands",
	Description: `
Run an app specific command.

These commands are bash functions, defined in the abra.sh of the recipe itself.
They can be run within the context of a service (e.g. app) or locally on your
work station by passing "--local". Arguments can be passed into these functions
using the "-- <args>" syntax.

Example:

  abra app cmd example.com app create_user -- me@example.com
`,
	ArgsUsage: "<domain> [<service>] <command> [-- <args>]",
	Flags: []cli.Flag{
		internal.DebugFlag,
		internal.LocalCmdFlag,
		internal.RemoteUserFlag,
		internal.TtyFlag,
		internal.OfflineFlag,
	},
	BashComplete: autocomplete.AppNameComplete,
	Before:       internal.SubCommandBefore,
	Action: func(c *cli.Context) error {
		conf := runtime.New(runtime.WithOffline(internal.Offline))
		app := internal.ValidateApp(c, conf)

		cl, err := client.New(app.Server)
		if err != nil {
			logrus.Fatal(err)
		}

		if internal.LocalCmd && internal.RemoteUser != "" {
			internal.ShowSubcommandHelpAndError(c, errors.New("cannot use --local & --user together"))
		}

		hasCmdArgs, parsedCmdArgs := parseCmdArgs(c.Args(), internal.LocalCmd)

		abraSh := path.Join(config.RECIPES_DIR, app.Recipe, "abra.sh")
		if _, err := os.Stat(abraSh); err != nil {
			if os.IsNotExist(err) {
				logrus.Fatalf("%s does not exist for %s?", abraSh, app.Name)
			}
			logrus.Fatal(err)
		}

		if internal.LocalCmd {
			cmdName := c.Args().Get(1)
			if err := internal.EnsureCommand(abraSh, app.Recipe, cmdName); err != nil {
				logrus.Fatal(err)
			}

			logrus.Debugf("--local detected, running %s on local work station", cmdName)

			var exportEnv string
			for k, v := range app.Env {
				exportEnv = exportEnv + fmt.Sprintf("%s='%s'; ", k, v)
			}
			var sourceAndExec string
			if hasCmdArgs {
				logrus.Debugf("parsed following command arguments: %s", parsedCmdArgs)
				sourceAndExec = fmt.Sprintf("TARGET=local; APP_NAME=%s; STACK_NAME=%s; %s . %s; %s %s", app.Name, app.StackName(), exportEnv, abraSh, cmdName, parsedCmdArgs)
			} else {
				logrus.Debug("did not detect any command arguments")
				sourceAndExec = fmt.Sprintf("TARGET=local; APP_NAME=%s; STACK_NAME=%s; %s . %s; %s", app.Name, app.StackName(), exportEnv, abraSh, cmdName)
			}

			shell := "/bin/bash"
			if _, err := os.Stat(shell); errors.Is(err, os.ErrNotExist) {
				logrus.Debugf("%s does not exist locally, use /bin/sh as fallback", shell)
				shell = "/bin/sh"
			}
			cmd := exec.Command(shell, "-c", sourceAndExec)

			if err := internal.RunCmd(cmd); err != nil {
				logrus.Fatal(err)
			}
		} else {
			targetServiceName := c.Args().Get(1)

			cmdName := c.Args().Get(2)
			if err := internal.EnsureCommand(abraSh, app.Recipe, cmdName); err != nil {
				logrus.Fatal(err)
			}

			serviceNames, err := config.GetAppServiceNames(app.Name)
			if err != nil {
				logrus.Fatal(err)
			}

			matchingServiceName := false
			for _, serviceName := range serviceNames {
				if serviceName == targetServiceName {
					matchingServiceName = true
				}
			}

			if !matchingServiceName {
				logrus.Fatalf("no service %s for %s?", targetServiceName, app.Name)
			}

			logrus.Debugf("running command %s within the context of %s_%s", cmdName, app.StackName(), targetServiceName)

			if hasCmdArgs {
				logrus.Debugf("parsed following command arguments: %s", parsedCmdArgs)
			} else {
				logrus.Debug("did not detect any command arguments")
			}

			if err := internal.RunCmdRemote(cl, app, abraSh, targetServiceName, cmdName, parsedCmdArgs); err != nil {
				logrus.Fatal(err)
			}
		}

		return nil
	},
}

func parseCmdArgs(args []string, isLocal bool) (bool, string) {
	var (
		parsedCmdArgs string
		hasCmdArgs    bool
	)

	if isLocal {
		if len(args) > 2 {
			return true, fmt.Sprintf("%s ", strings.Join(args[2:], " "))
		}
	} else {
		if len(args) > 3 {
			return true, fmt.Sprintf("%s ", strings.Join(args[3:], " "))
		}
	}

	return hasCmdArgs, parsedCmdArgs
}
