package app

import (
	"context"
	"fmt"
	"strings"

	"coopcloud.tech/abra/cli/internal"
	"coopcloud.tech/abra/pkg/autocomplete"
	"coopcloud.tech/abra/pkg/client"
	"coopcloud.tech/abra/pkg/formatter"
	"coopcloud.tech/abra/pkg/runtime"
	"coopcloud.tech/abra/pkg/service"
	stack "coopcloud.tech/abra/pkg/upstream/stack"
	"github.com/docker/docker/api/types"
	"github.com/sirupsen/logrus"
	"github.com/urfave/cli"
)

var appServicesCommand = cli.Command{
	Name:      "services",
	Aliases:   []string{"sr"},
	Usage:     "Display all services of an app",
	ArgsUsage: "<domain>",
	Flags: []cli.Flag{
		internal.DebugFlag,
		internal.OfflineFlag,
	},
	Before:       internal.SubCommandBefore,
	BashComplete: autocomplete.AppNameComplete,
	Action: func(c *cli.Context) error {
		conf := runtime.New(runtime.WithOffline(internal.Offline))
		app := internal.ValidateApp(c, conf)

		cl, err := client.New(app.Server)
		if err != nil {
			logrus.Fatal(err)
		}

		isDeployed, _, err := stack.IsDeployed(context.Background(), cl, app.StackName())
		if err != nil {
			logrus.Fatal(err)
		}

		if !isDeployed {
			logrus.Fatalf("%s is not deployed?", app.Name)
		}

		filters, err := app.Filters(true, true)
		if err != nil {
			logrus.Fatal(err)
		}

		containers, err := cl.ContainerList(context.Background(), types.ContainerListOptions{Filters: filters})
		if err != nil {
			logrus.Fatal(err)
		}

		tableCol := []string{"service name", "image"}
		table := formatter.CreateTable(tableCol)

		for _, container := range containers {
			var containerNames []string
			for _, containerName := range container.Names {
				trimmed := strings.TrimPrefix(containerName, "/")
				containerNames = append(containerNames, trimmed)
			}

			serviceShortName := service.ContainerToServiceName(container.Names, app.StackName())
			serviceLongName := fmt.Sprintf("%s_%s", app.StackName(), serviceShortName)

			tableRow := []string{
				serviceLongName,
				formatter.RemoveSha(container.Image),
			}
			table.Append(tableRow)
		}

		table.Render()

		return nil
	},
}
